/*
 * MyDocument.m
 * Ring Tones
 *
 * Created by Graham Booker on Fri Jun 13 2003.
 * Copyright (c) 2003 Graham Booker.
 * All rights reserved.
 *
 *  Permission is hereby granted, free of charge, to any person obtaining
 *  a copy of this software and associated documentation files (the "Software"),
 *  to deal in the Software without restriction, including without limitation
 *  the rights to use, copy, modify, merge, publish, distribute, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, provided that the above copyright notice(s) and this
 *  permission notice appear in all copies of the Software and that both the
 *  above copyright notice(s) and this permission notice appear in supporting
 *  documentation.
 *
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT OF THIRD PARTY RIGHTS.
 *  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR HOLDERS INCLUDED IN THIS NOTICE BE
 *  LIABLE FOR ANY CLAIM, OR ANY SPECIAL INDIRECT OR CONSEQUENTIAL DAMAGES, OR
 *  ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
 *  IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
 *  OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *
 *  Except as contained in this notice, the name of a copyright holder shall not
 *  be used in advertising or otherwise to promote the sale, use or other dealings
 *  in this Software without prior written authorization of the copyright holder. 
 */

#import "MyDocument.h"

@implementation MyDocument

- (id)init
{
    self = [super init];
    if (self) {
    
        // Add your subclass-specific initialization here.
        // If an error occurs here, send a [self release] message and return nil.

    }
    return self;
}

- (NSString *)windowNibName
{
    // Override returning the nib file name of the document
    // If you need to use a subclass of NSWindowController or if your document supports multiple NSWindowControllers, you should remove this method and override -makeWindowControllers instead.
    return @"MyDocument";
}

- (void)windowControllerDidLoadNib:(NSWindowController *) aController
{
    [super windowControllerDidLoadNib:aController];
    // Add any code here that needs to be executed once the windowController has loaded the document's window.
    [statusField setStringValue:@"Not Ready"];
    [skipField setEnabled:NO];
    [durationField setEnabled:NO];
}

- (NSData *)dataRepresentationOfType:(NSString *)aType
{
    // Insert code here to write your document from the given data.  You can also choose to override -fileWrapperRepresentationOfType: or -writeToFile:ofType: instead.
    return nil;
}

- (BOOL)loadDataRepresentation:(NSData *)data ofType:(NSString *)aType
{
    // Insert code here to read your document from the given data.  You can also choose to override -loadFileWrapperRepresentation:ofType: or -readFromFile:ofType: instead.
    return YES;
}

- (IBAction)changeInput:(id)sender{
    NSArray *fileTypes = [NSArray arrayWithObjects:@"mp3", @"MP3", nil];
    NSOpenPanel *oPanel = [NSOpenPanel openPanel];
    NSString *input = [inputField stringValue];
    int result;

    [oPanel setAllowsMultipleSelection:NO];
    [oPanel setTreatsFilePackagesAsDirectories:YES];
    [oPanel setCanChooseDirectories:NO];
    result = [oPanel runModalForDirectory:[input stringByDeletingLastPathComponent] file:[input lastPathComponent] types:fileTypes];

    if (result == NSOKButton) {
        input = [[oPanel filename] retain];
        [inputField setStringValue:input];
        if([[outputField stringValue] length] == 0){
            [outputField setStringValue:[[input stringByDeletingPathExtension] stringByAppendingPathExtension:@"pdb"]];
        }
        if([[ringnameField stringValue] length] == 0){
            [ringnameField setStringValue:[[input lastPathComponent] stringByDeletingPathExtension]];
        }
    }
    [self checkstatus:sender];
}

- (IBAction)changeOutput:(id)sender{
    NSSavePanel *savePanel = [NSSavePanel savePanel];
    NSString *destination = [outputField stringValue];

    [savePanel setRequiredFileType:@"pdb"];
    if([savePanel runModalForDirectory:[destination stringByDeletingLastPathComponent] file:[destination lastPathComponent]] == NSOKButton){
        destination = [[savePanel filename] retain];
        [outputField setStringValue:destination];
    }
    [self checkstatus:sender];
}

BOOL valid(mpegheader_t head){
    if(head.sync != 0x7ff){
        return NO;
    }
    if(head.version == 1){
        return NO;
    }
    if(head.layer == 0){
        return NO;
    }
    if(head.bitrate == 0){
        return NO;
    }
    if(head.bitrate == 0xf){
        return NO;
    }
    if(head.samplerate == 0x3){
        return NO;
    }
    return YES;
}

static int samplerates[3][4]={
    {44100, 48000, 32000},
    {22050, 24000, 16000},
    {11025, 12000, 8000}
};

int length(mpegheader_t head){
    static int bitrates[3][3][16]={{
        {0, 32, 64, 96, 128, 160, 192, 224, 256, 288, 320, 352, 384, 416, 448, 0},
        {0, 32, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, 384, 0},
        {0, 32, 40, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, 0}},
    {   {0, 32, 48, 56, 64, 80, 96, 112, 128, 144, 160, 192, 224, 256, 320, 0},
        {0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160, 0},
        {0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160, 0}},
    {   {0, 32, 48, 56, 64, 80, 96, 112, 128, 144, 160, 192, 224, 256, 320, 0},
        {0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160, 0},
        {0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160, 0},
    }};

    int layer = 3-head.layer;
    int version = 0;

    switch(head.version){
        case 0:
            version = 2;
            break;
        case 2:
            version = 1;
            break;
        case 3:
            version = 0;
    }

    if(layer == 0){
        return (12 * bitrates[version][layer][head.bitrate]*1000 / samplerates[version][head.samplerate] + head.padding) * 4;
    } else {
        return 144 * bitrates[version][layer][head.bitrate]*1000 / samplerates[version][head.samplerate] + head.padding;
    }
}

int setLimits(FILE *in, int totsize, int limitsize, float skip, float duration){
    long lastpos = 0;
    int i, seek = 0, framecount;
    BOOL sizelimit = limitsize != 0x7fffffff;
    int skipframecount, durationcount;
    int version = 0;
    mpegheader_t head;
    
    for(i=0;i<totsize;i++){
        unsigned char buff[4];
        int validcount=0;

        fread(buff, 1, 1, in);
        if(buff[0] == 0xff){
            lastpos = ftell(in);
            fread(buff+1, 1, 3, in);
        }
        memcpy(&head, buff, 4);
        while((valid(head)) && (validcount < 4)){
            validcount++;
            fseek(in, length(head)-4, SEEK_CUR);
            fread(&head, 1, 4, in);
        }
        if(validcount != 4){
            fseek(in, lastpos, SEEK_SET);
        }else{
            fseek(in, lastpos-1, SEEK_SET);
            break;
        }
    }
    switch(head.version){
        case 0:
            version = 2;
            break;
        case 2:
            version = 1;
            break;
        case 3:
            version = 0;
    }
    if(head.layer == 0x3){
        skipframecount = samplerates[version][head.samplerate] / 384 * skip;
        durationcount = skipframecount + samplerates[version][head.samplerate] / 384 * duration;
    } else {
        skipframecount = samplerates[version][head.samplerate] / 1152 * skip;
        durationcount = skipframecount + samplerates[version][head.samplerate] / 1152 * duration;
    }
    framecount = 0;
    while(i<totsize){
        int readcount;
        mpegheader_t head;

        fread(&head, 1, 4, in);
        if(!valid(head)){
            break;
        }
        readcount = length(head);
        if(sizelimit){
            if(readcount + i +128 > limitsize){
                break;
            }
        } else {
            if(framecount == skipframecount){
                seek=i;
            }
            if(framecount == durationcount){
                break;
            }
        }
        fseek(in, readcount-4, SEEK_CUR);
        i+= readcount;
        framecount++;
    }
    fseek(in, seek, SEEK_SET);
    return i-seek;
}

- (IBAction)runConvert:(id)sender{
    static char types[9][5] = {
        "ring", "syst", "demo",
        "tone", "wild", "snde",
        "alrm", "flip", "spch"
    };
    NSString *ringname, *inputname, *outputname;
    pdbheader_t header;
    FILE *in, *out;
    struct stat filestat;
    int filesize, i, offset, limitsize=0x7fffffff;
    float skip=0, duration=0;
    short recordcount;
    recordheader_t *records;
    id3tag_t tag;

    ringname = [ringnameField stringValue];
    inputname = [inputField stringValue];
    outputname = [outputField stringValue];
    if([ringname length] > 31){
        NSBeginAlertSheet(@"Error", @"OK", nil, nil, theWindow, self, nil, NULL, nil, @"Ring name must not be more than 31 characters", nil);
        return;
    }
    if([ringname length] == 0){
        NSBeginAlertSheet(@"Error", @"OK", nil, nil, theWindow, self, nil, NULL, nil, @"Please set a ring name", nil);
    return;
    }
    if((in = fopen([inputname cString], "r")) == NULL){
        NSBeginAlertSheet(@"Error", @"OK", nil, nil, theWindow, self, nil, NULL, nil, @"Couldn't not open input file", nil);
        return;
    }
    if((out = fopen([outputname cString], "w")) == NULL){
        NSBeginAlertSheet(@"Error", @"OK", nil, nil, theWindow, self, nil, NULL, nil, @"Couldn't write to output path", nil);
        return;
    }

    if([limitButtons selectedRow] == 0){
        if([[limitField stringValue] length] != 0){
            double length = [limitField doubleValue];
            if(length == 0){
                NSBeginAlertSheet(@"Error", @"OK", nil, nil, theWindow, self, nil, NULL, nil, @"Limit Size is not valid", nil);
                return;
            }
            limitsize = length*1000;
        }
    } else {
        skip = [skipField doubleValue];
        if(skip == 0){
            NSBeginAlertSheet(@"Error", @"OK", nil, nil, theWindow, self, nil, NULL, nil, @"Skip time is not valid", nil);
            return;
        }
        duration = [durationField doubleValue];
        if(duration == 0){
            NSBeginAlertSheet(@"Error", @"OK", nil, nil, theWindow, self, nil, NULL, nil, @"Duration time is not valid", nil);
            return;
        }
    }

    [statusField setStringValue:@"Creating Header"];
    stat([inputname cString], &filestat);
    memset(&header, '\0', sizeof(pdbheader_t));
    strcpy(header.name, [ringname cString]);
    header.setto20 = 0x20;
    memcpy(header.ring, types[[typeButtons selectedRow]*3+[typeButtons selectedColumn]], 4);
    memcpy(header.MP3P, "MP3P", 4);
    filesize = filestat.st_size;
    if((limitsize < filesize) || (duration != 0)){
        filesize = setLimits(in, filesize, limitsize, skip, duration);
    }
    recordcount = (filesize -1 + RECORDSIZE)/RECORDSIZE;
    header.recordcount = recordcount;

    [statusField setStringValue:@"Writing Records"];
    records = (recordheader_t *)malloc(recordcount*sizeof(recordheader_t));
    offset = sizeof(pdbheader_t)+recordcount*sizeof(recordheader_t);
    for(i=0;i<recordcount;i++){
        records[i].recordoffset = offset;
        records[i].unknown = 0;
        offset += RECORDSIZE;
    }
    fwrite(&header, 1, sizeof(pdbheader_t), out);
    fwrite(records, sizeof(recordheader_t), recordcount, out);
    [statusField setStringValue:@"Copying MP3 Data"];
    filesize-=sizeof(id3tag_t);
    for(i=0;i<filesize;){
        char buffer[4096];
        int readcount=4096;

        if(filesize-i < 4096){
            readcount = filesize-i;
        }

        readcount = fread(buffer, 1, readcount, in);
        fwrite(buffer, 1, readcount, out);
        i+=readcount;
    }
    [statusField setStringValue:@"Writing the strange ID3 Tag"];
    memset(&tag, '\0', sizeof(id3tag_t));
    memcpy(tag.name, "TAG", 3);
    tag.year = 0x30000000;
    tag.genre = 0xff;
    fwrite(&tag, sizeof(id3tag_t), 1, out);
    fclose(in);
    fclose(out);
    [statusField setStringValue:@"Done"];
}

- (IBAction)checkstatus:(id)sender{
    if([[inputField stringValue] length] == 0){
        [statusField setStringValue:@"Input File Unspecified"];
        return;
    }
    if([[outputField stringValue] length] == 0){
        [statusField setStringValue:@"Output File Unspecified"];
        return;
    }
    if([[ringnameField stringValue] length] == 0){
        [statusField setStringValue:@"Ring Name Unspecified"];
        return;
    }
    if([[ringnameField stringValue] length] > 31){
        [statusField setStringValue:@"Ring Name too long"];
        return;
    }
    if([limitButtons selectedRow] == 0){
        if([[limitField stringValue] length] != 0){
            double length = [limitField doubleValue];
            if(length == 0){
                [statusField setStringValue:@"Limit Size is not valid"];
                return;
            }
        }
    } else {
        double skip = [skipField doubleValue];
        double duration = [durationField doubleValue];
        if(skip == 0){
            [statusField setStringValue:@"Skip time is not valid"];
            return;
        }
        if(duration == 0){
            [statusField setStringValue:@"Duration time is not valid"];
            return;
        }
    }
    [statusField setStringValue:@"Ready"];    
}

- (IBAction)changeToSize:(id)sender{
    [skipField setEnabled:NO];
    [durationField setEnabled:NO];
    [limitField setEnabled:YES];
    [self checkstatus:sender];
}

- (IBAction)changeToTime:(id)sender{
    [skipField setEnabled:YES];
    [durationField setEnabled:YES];
    [limitField setEnabled:NO];
    [self checkstatus:sender];
}

@end